import IKCache from '../Prototype/IKCache';
export default class IKCache_Redis extends IKCache {
  constructor(config, transaction) {
    super(config, transaction);
    //
    this.awaitingConnectionQueue = null;
    //
    const localConsole = (transaction ? transaction.logger : console);
    if (config && config.host) { localConsole.debug(`Using cache: REDIS on host: ${config.host}`); }
  }
  /* Public */
  async get(key) {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.get(key, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async set(key, value) {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.set(key, value, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async del(key) {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.del(key, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async keys(pattern) {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.keys(pattern, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async mget() {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.mget(...arguments, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async mset() {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.mset(...arguments, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  async mdel() {
    return new Promise(async (resolve, reject) => {
      await this._connect();
      this.connection.del(...arguments, (err, reply) => {
        if (err) reject(err);
        else (resolve(reply));
      });
    });
  }
  /* Private */
  _isRedisClientAvailable() {
    let response = true;
    try { const test = require('redis'); } 
    catch (e) { response = false; }
    return response;
  }
  _isRedisClientV4() {
    let isV4 = false;
    try {  isV4 = !!require('redis').createCluster; } 
    catch (e) { }
    return isV4;
  }
  async _connect() {
    //No connection, but waiting connection queue is valid? wait until other promise
    //fulfill this promire
    if (!this.connection && this.awaitingConnectionQueue) {
      return new Promise((resolve) => this.awaitingConnectionQueue.push(resolve));
    } else if (this.connection) return this.connection; //already connected

    //Check if redis client is available
    if (!this._isRedisClientAvailable()) throw new Error('Redis client is not available! Please, run "npm i -S redis".');

    //Prepare
    const localConsole = (this.transaction ? this.transaction.logger : console);
    this.awaitingConnectionQueue = [];
    localConsole.debug("Starting remote cache connection");
    
    //Instantiate client and initiate connection
    return new Promise(async (resolve, reject) => {
      const redis = require("redis");
      const isV4 = this._isRedisClientV4();
      //
      const connection = redis.createClient({
        ...(isV4 ? {
          username: this.config.user, 
          socket: { 
            ...(this.config.enableTLS ? { tls: true } : {}), host: this.config.host
          }
        } : {
          host: this.config.host, user: this.config.user,
          ...(this.config.enableTLS ? { tls: {} } /* https://docs.upstash.com/howto/connectwithtls */ : {})
        }),
        password: this.config.password, 
      });
      //redis v4 & v3 compatibility
      if (connection && connection.connect) await connection.connect();
      //
      const connectedHandler = () => {
        this.connection = connection;
        if (this.awaitingConnectionQueue) this.awaitingConnectionQueue.forEach((resolve) => resolve());
        this.awaitingConnectionQueue = null;
        resolve();
      }
      //reached here on v4, everything is connected
      if (isV4) {
        connectedHandler();
        return;
      } 
      //v3 <=
      connection.on('connect', connectedHandler);
      connection.on('error', (err) => {
        localConsole.error(err);
        if (this.awaitingConnectionQueue) this.awaitingConnectionQueue.forEach((resolve) => resolve()); //fulfill, error will be thrown
        this.awaitingConnectionQueue = null;
        reject(err);
      });
    });
  }
}
